// Project Name: OpenMEEG (http://openmeeg.github.io)
// © INRIA and ENPC under the French open source license CeCILL-B.
// See full copyright notice in the file LICENSE.txt
// If you make a copy of this file, you must either:
// - provide also LICENSE.txt and modify this header to refer to it.
// - replace this header by the LICENSE.txt content.

#pragma once

#include <iostream>
#include <fstream>

#include <map>
#include <string>

#include <om_utils.h>
#include <MeshIO.h>
#include <OMExceptions.H>

#ifdef USE_VTK
#include <vtkPolyData.h>
#include <vtkPoints.h>
#include <vtkPolyDataReader.h>
#include <vtkSmartPointer.h>
#include <vtkXMLPolyDataReader.h>
#include <vtkDataReader.h>
#include <vtkCellArray.h>
#include <vtkCharArray.h>
#include <vtkPointData.h>
#include <vtkDataArray.h>
#endif

namespace OpenMEEG::MeshIOs {

    /// \brief Mesh io for VTK file format.

    class OPENMEEG_EXPORT Vtk: public MeshIO {

        typedef MeshIO base;

    public:

    #ifdef USE_VTK

        ~Vtk() {
            if (buffer)
                delete[] buffer;
        }

        void load_points(Geometry& geom) override {
            //  TODO: Replace with newer filesystem calls.
            fs.seekg (0,ios::end);
            const unsigned length = fs.tellg();
            fs.seekg (0,ios::beg);

            // Read data as a block and pass it to vtk.

            buffer = new char[length];
            fs.read(buffer,length);

            vtkSmartPointer<vtkCharArray> buf = vtkSmartPointer<vtkCharArray>::New();
            buf->SetArray(buffer,length,1);

            //  Create the vtk reader.
            //  Make it read from the InputArray 'buf' instead of the default.

            vtkSmartPointer<vtkPolyDataReader> reader = vtkSmartPointer<vtkPolyDataReader>::New();
            reader->SetInputArray(buf);
            reader->SetReadFromInputString(1);
            if (!reader->IsFilePolyData()) {
                reader->Delete();
                throw OpenMEEG::VTKError("Mesh \"" + fname + "\" is not a valid vtk poly data file");
            }

            reader->Update();
            vtkMesh = reader->GetOutput();

            const unsigned npts = vtkMesh->GetNumberOfPoints();
            Vertices vertices;
            for (unsigned i=0;i<npts;++i)
                vertices.push_back(Vertex(vtkMesh->GetPoint(i)));
            indmap = geom.add_vertices(vertices);
        }

        void load_triangles(OpenMEEG::Mesh& mesh) override {
            reference_vertices(mesh);

            const unsigned ntrgs = vtkMesh->GetNumberOfCells();
            mesh.triangles().reserve(ntrgs);

            for (unsigned i = 0; i < ntrgs; ++i) {
                if (vtkMesh->GetCellType(i)!=VTK_TRIANGLE)
                    throw OpenMEEG::VTKError("Mesh \"" + fname + "\" is not a triangulation");
                vtkIdList* l = vtkMesh->GetCell(i)->GetPointIds();
                mesh.add_triangle(TriangleIndices(l->GetId(0),l->GetId(1),l->GetId(2)),indmap);
            }
        }
    #else
        static unsigned vtk_error() {
            throw OpenMEEG::VTKError("OpenMEEG was not compiled with VTK support. Specify USE_VTK in cmake.");
        }

        void     load_points(Geometry&)          override { vtk_error(); }
        void     load_triangles(OpenMEEG::Mesh&) override { vtk_error(); }
        void     load(OpenMEEG::Mesh&)           override { vtk_error(); }
    #endif

        void save(const OpenMEEG::Mesh& mesh,std::ostream& os) const override {
            os << "# vtk DataFile Version 2.0" << std::endl;
            os << "Mesh file generated by OpenMEEG" << std::endl;
            os << "ASCII" << std::endl;
            os << "DATASET POLYDATA" << std::endl;
            os << "POINTS " << mesh.vertices().size() << " float" << std::endl;

            const VertexIndices& vertex_index(mesh);
            for (const auto& vertex : mesh.vertices())
                os << *vertex << std::endl;

            os << "POLYGONS " << mesh.triangles().size() << ' ' << mesh.triangles().size()*4 << std::endl;
            for (const auto& triangle : mesh.triangles())
                os << "3 " << vertex_index(triangle,0) << ' '
                           << vertex_index(triangle,1) << ' '
                           << vertex_index(triangle,2) << std::endl;

            os << "CELL_DATA " << mesh.triangles().size() << std::endl;
            os << "POINT_DATA " << mesh.vertices().size() << std::endl;
            os << "NORMALS normals float" << std::endl;
            for (const auto& vertex : mesh.vertices())
                os << mesh.normal(*vertex) << std::endl;
        }

        MeshIO* clone(const std::string& filename) const override { return new Vtk(filename); }

    private:

        Vtk(const std::string& filename=""): base(filename,"vtk") { }

        #ifdef USE_VTK
        char*                        buffer = nullptr;
        vtkSmartPointer<vtkPolyData> vtkMesh;
        #endif

        static const Vtk prototype;

        const char* name() const override { return "VTK"; }
    };
}
